@@ -403,9 +403,9 @@ sub convert_value_to_cpp($)
 {
   my ($text) = @_;
 
-  $$text =~ s"\bFALSE\b"<tt>false</tt>"g;
-  $$text =~ s"\bTRUE\b"<tt>true</tt>"g;
-  $$text =~ s"\bNULL\b"<tt>nullptr</tt>"g;
+  $$text =~ s"`?\bFALSE\b`?"<tt>false</tt>"g;
+  $$text =~ s"`?\bTRUE\b`?"<tt>true</tt>"g;
+  $$text =~ s"`?\bNULL\b`?"<tt>nullptr</tt>"g;
 
   # Enumerator names
   $$text =~ s/\b([A-Z]+)_([A-Z\d_]+)\b/&DocsParser::substitute_enumerator_name($1, $2)/eg;
@@ -423,6 +423,9 @@ sub remove_example_code($$)
   $example_removals +=
     ($$text =~ s"<programlisting>.*?</programlisting>"\n[C example ellipted]"sg);
   $example_removals += ($$text =~ s"\|\[.*?]\|"\n[C example ellipted]"sg);
+  # gi-docgen syntax.
+  # remove_example_code() is called after add_m4_quotes().
+  $example_removals += ($$text =~ s"(?:'__BT__`){3}[cC].*?(?:'__BT__`){3}"\n[C example ellipted]"sg);
 
   # See "MS Visual Studio" comment in gmmproc.in.
   print STDERR "gmmproc, $main::source, $obj_name: Example code discarded.\n"
@@ -698,11 +701,14 @@ sub convert_tags_to_doxygen($)
     # gtk-doc uses @thearg, but doxygen uses @a thearg.
     s" ?\@([a-zA-Z0-9]*(_[a-zA-Z0-9]+)*)_?\b" \@a $1"g;
 
-    # Don't convert Doxygen's $throw, @throws and @param, so these can be used
+    # Don't convert Doxygen's @throw, @throws and @param, so these can be used
     # in the docs_override.xml.
     # Also don't convert @enum and @var which are used for enum documentation.
     s" \@a (throws?|param|enum|var)\b" \@$1"g;
 
+    # gi-docgen uses [type@Module.*]. Don't convert them.
+    s"(\[[a-z]+?) \@a (.+?])"$1\@$2"g;
+
     s"^Note ?\d?: "\@note "mg;
     s"</?programlisting>""g;
     s"<!>""g;
@@ -738,9 +744,9 @@ sub convert_tags_to_doxygen($)
     # Doxygen is too dumb to handle &mdash;
     s"&mdash;" \@htmlonly&mdash;\@endhtmlonly "g;
 
-    s"\%?\bFALSE\b"<tt>false</tt>"g;
-    s"\%?\bTRUE\b"<tt>true</tt>"g;
-    s"\%?\bNULL\b"<tt>nullptr</tt>"g;
+    s"`?\%?\bFALSE\b`?"<tt>false</tt>"g;
+    s"`?\%?\bTRUE\b`?"<tt>true</tt>"g;
+    s"`?\%?\bNULL\b`?"<tt>nullptr</tt>"g;
 
     s"#?\bgboolean\b"<tt>bool</tt>"g;
     s"#?\bg(int|short|long)\b"<tt>$1</tt>"g;
@@ -823,6 +829,9 @@ sub substitute_identifiers($$)
   {
     # TODO: handle more than one namespace
 
+    # gi-docgen syntax for links to symbols is described at
+    # https://gnome.pages.gitlab.gnome.org/gi-docgen/linking.html
+
     # Convert property names to C++.
     # The standard (and correct) gtk-doc way of referring to properties.
     s/(#[A-Z]\w+):([a-z\d-]+)/my $name = "$1::property_$2()"; $name =~ s"-"_"g; "$name";/ge;
@@ -838,16 +847,22 @@ sub substitute_identifiers($$)
     # but a few are written with the double colon in the gtk+ docs so this
     # protects against those errors.
     s/([A-Z]\w+)::([a-z\d-]+)(\s+property)/my $name = "$1::property_$2()$3"; $name =~ s"-"_"g; "$name";/ge;
+    # gi-docgen syntax.
+    s/\[(`?)property@(?:([A-Z]\w*)\.)?([A-Z]\w+):([a-z\d-]+)\1]/"$1" . &DocsParser::substitute_property_or_signal_name($doc_func, $2, $3, "property", $4) . "$1"/ge;
 
     # Convert signal names to C++.
     s/(^|\s)::([a-z\d-]+)(\(\))*([^:\w]|$)/my $name = "$1signal_$2()$4"; $name =~ s"-"_"g; "$name";/ge;
     s/(#[A-Z]\w+)::([a-z\d-]+)(\(\))*([^:\w]|$)/my $name = "$1::signal_$2()$4"; $name =~ s"-"_"g; "$name";/ge;
+    s/\[(`?)signal@(?:([A-Z]\w*)\.)?([A-Z]\w+)::([a-z\d-]+)\1]/"$1" . &DocsParser::substitute_property_or_signal_name($doc_func, $2, $3, "signal", $4) . "$1"/ge;
 
     # Type names
-    s/[#%]([A-Z][a-z]*)([A-Z][A-Za-z]+)\b/&DocsParser::substitute_type_name($1, $2)/eg;
+    s/[#%]([A-Z][a-z]*)([A-Z][A-Za-z]+)\b/&DocsParser::substitute_type_name($doc_func, $1, $2)/eg;
+    s/`([A-Z][a-z]*)([A-Z][A-Za-z]*[a-z])`/"`" . &DocsParser::substitute_type_name($doc_func, $1, $2) . "`"/eg;
+    s/\[(`?)(?:class|enum|error|flags|iface|struct|type)@(?:([A-Z]\w*)\.)?([A-Z]\w+)\1]/"$1" . &DocsParser::substitute_type_name($doc_func, $2, $3) . "$1"/eg;
 
     # Enumerator names
     s/[#%]([A-Z]+)_([A-Z\d_]+)\b/&DocsParser::substitute_enumerator_name($1, $2)/eg;
+    s/`([A-Z]+)_([A-Z\d_]+)`/"`" . &DocsParser::substitute_enumerator_name($1, $2) . "`"/eg;
 
     s/\bG:://g; #Rename G::Something to Something.
 
@@ -856,12 +871,18 @@ sub substitute_identifiers($$)
 
     # Replace C function names with C++ counterparts.
     s/\b([a-z]+_[a-z][a-z\d_]+) ?\(\)/&DocsParser::substitute_function($doc_func, $1)/eg;
+    s/\[(`?)id@([a-z\d_]+)\1]/"$1" . &DocsParser::substitute_function($doc_func, $2) . "$1"/eg;
+    s/\[(`?)(?:ctor|method)@(?:([A-Z]\w*)\.)?([A-Z]\w+)\.([a-z\d_]+)\1]/"$1" . &DocsParser::substitute_split_function($doc_func, $2, $3, $4) . "$1"/eg;
+    s/\[(`?)vfunc@(?:([A-Z]\w*)\.)?([A-Z]\w+)\.([a-z\d_]+)\1]/"$1" . &DocsParser::substitute_split_function($doc_func, $2, $3, $4 . "_vfunc") . "$1"/eg;
+    s/\[(`?)func@([\w.]+)\1]/"$1" . &DocsParser::substitute_func_function($doc_func, $2) . "$1"/eg;
   }
 }
 
-sub substitute_type_name($$)
+sub substitute_type_name($$$)
 {
-  my ($module, $name) = @_;
+  my ($doc_func, $module, $name) = @_;
+  $module = get_module_from_doc_func($doc_func) if !$module;
+
   my $c_name = $module . $name;
 
   if (exists $DocsParser::type_names{$c_name})
@@ -911,7 +932,7 @@ sub substitute_enumerator_name($$)
 
   # Don't apply the default substitution to these module names.
   # They are not really modules.
-  if (grep {$module eq $_} qw(HAS NO O SO AF))
+  if (grep {$module eq $_} qw(HAS NO O SO AF XDG))
   {
     return $c_name;
   }
@@ -923,6 +944,79 @@ sub substitute_enumerator_name($$)
   return $cxx_name;
 }
 
+sub substitute_property_or_signal_name($$$$$)
+{
+  # $doc_func can be the name of a property (ModuleClass:property_name),
+  # signal (ModuleClass::signal_name), function/method (module_class_method_name)
+  # or class/enum/etc. (ModuleClass).
+  my ($doc_func, $module, $class, $prop_or_sig, $name) = @_;
+  $module = get_module_from_doc_func($doc_func) if !$module;
+
+  my $prefix = $module . $class;
+  $name =~ s"-"_"g;
+  $name = $prop_or_sig . "_" . $name . "()";
+  if (index($doc_func, $prefix . ":") == 0)
+  {
+    # Documentation of property or signal in the same class as the referred
+    # property or signal.
+    return $name;
+  }
+  if (index($doc_func, ":") == -1)
+  {
+    # Documentation of a function or method.
+    if (my $defs_method = GtkDefs::lookup_method_dont_mark($doc_func))
+    {
+      if ($$defs_method{class} eq $prefix)
+      {
+        # Documentation of function/method in the same class as the referred
+        # property or signal.
+        return $name;
+      }
+    }
+  }
+  return $module . "::" . $class . "::" . $name;
+}
+
+sub substitute_split_function($$$$)
+{
+  my ($doc_func, $module, $class, $name) = @_;
+  $module = get_module_from_doc_func($doc_func) if !$module;
+
+  my $prefix = build_method_prefix($module, $class);
+
+  if ($doc_func =~ m/^$prefix/)
+  {
+    return $name . "()";
+  }
+  else
+  {
+    return $module . "::" . $class . "::" . $name . "()";
+  }
+}
+
+sub substitute_func_function($$)
+{
+  # $name == Gtk.WidgetPaintable.func or Gtk.func or func.
+  my ($doc_func, $name) = @_;
+  $name =~ s/\./_/g;
+  $name =~ s/([a-z])([A-Z])/$1_$2/g;
+  return DocsParser::substitute_function($doc_func, lc($name));
+}
+
+sub get_module_from_doc_func($)
+{
+  my ($doc_func) = @_;
+
+  if ($doc_func =~ /^([a-z]+)_/)
+  {
+    # Function name. gtk_foo_bar -> Gtk
+    return "\u$1";
+  }
+  # Class name. GtkFooBar -> Gtk
+  $doc_func =~ /^([A-Z][a-z]*?)[A-Z]/;
+  return $1;
+}
+
 sub substitute_function($$)
 {
   my ($doc_func, $name) = @_;
@@ -990,15 +1084,11 @@ sub lookup_object_of_method($$)
   return undef;
 }
 
-
 sub build_method_name($$$$)
 {
   my ($doc_func, $module, $class, $name) = @_;
 
-  my $prefix = $module . $class;
-
-  $prefix =~ s/([a-z])([A-Z])/$1_$2/g;
-  $prefix = lc($prefix) . '_';
+  my $prefix = build_method_prefix($module, $class);
 
   if($$name =~ m/^\Q$prefix\E/)
   {
@@ -1009,5 +1099,15 @@ sub build_method_name($$$$)
   }
 }
 
+sub build_method_prefix($$)
+{
+  my ($module, $class) = @_;
+
+  my $prefix = $module . $class;
+
+  $prefix =~ s/([a-z])([A-Z])/$1_$2/g;
+  $prefix =~ s/^(Gdk|Gtk)_GL([A-Z][a-z])/$1_GL_$2/; # Special cases, add an underline
+  return lc($prefix) . '_';
+}
 
 1; # indicate proper module load.
